package script;

import java.io.File;
import java.io.FileOutputStream;
import java.io.FileReader;
import java.io.IOException;
import java.io.OutputStreamWriter;
import java.util.HashMap;
import java.util.List;

import org.apache.jena.datatypes.RDFDatatype;
import org.apache.jena.datatypes.TypeMapper;
import org.apache.jena.ontology.Individual;
import org.apache.jena.ontology.OntClass;
import org.apache.jena.ontology.OntModel;
import org.apache.jena.ontology.OntModelSpec;
import org.apache.jena.rdf.model.ModelFactory;
import org.apache.jena.rdf.model.Property;
import org.apache.jena.rdf.model.Statement;
import org.apache.jena.rdf.model.StmtIterator;

import com.opencsv.CSVReader;

/**
 * Definici del script que poblar l'ontologia EVCP
 */
public class PopulateEVCPOntology {

	static String ontologyFile = "D:\\1..UOC_Patricia\\10e.Semestre\\2.TFG\\PACs\\PAC3\\0.Producte_TFG_v4\\workspace\\PopulateEVCPOntology\\src\\data\\ElectricVehicleChargingPointsOntology.owl";
	static String source = "http://www.semanticweb.org/patricia/ontologies/2019/1/ElectricVehicleChargingPointsOntology";
	static String NS = source + "#";
	static String geoNS = "http://www.w3.org/2003/01/geo/wgs84_pos" + "#";
	static String geosparqlNS = "http://www.opengis.net/ont/geosparql" + "#";
	static String CSVFileName = "D:\\1..UOC_Patricia\\10e.Semestre\\2.TFG\\PACs\\PAC3\\0.Producte_TFG_v4\\workspace\\PopulateEVCPOntology\\src\\data\\PUNTS_RECARREGA_VEHICLES_ELECTRICS.csv";
	
	
	/**
	 * Mtode principal (main) que  instancia un nou objecte de PopulateEVCPOntology
	 * i invoca al mtode d'aquest objecte per poblar l'EVCP (populate)
	 * @param argv (es podrien utilitzar per indicar el nom del CSV i de l'ontologia, per no ha sigut el cas)
	 * @throws IOException
	 *  llena una excepci en cas d'error durant el procs d'integraci de la informaci
	 *  del CSV dins l'EVCP
	 */
	public static void main(String argv[]) throws IOException{
		PopulateEVCPOntology evcp = new PopulateEVCPOntology();
		evcp.populate();
	}
	
	/**
	 * Mtode que llegeix el CSV i crea les instncies adients en l'EVCP.
	 * En finalitzar s'obtenen el fitxer de l'ontologia que inclou totes les instncies creades.
	 * Aquest fitxer s emmagatzemat en 3 formats: RDF, TTL i OWL
	 * 
	 * @throws IOException
	 *  llena una excepci en cas d'error durant el procs
	 */
	public void populate() throws IOException{
		OntModel model = ModelFactory.createOntologyModel(OntModelSpec.OWL_MEM);
		model.read(ontologyFile, "RDF/XML");
		System.out.println("NS: "+NS);
		System.out.println("-----------------------------");
		
		System.out.println("START: reading CSV...");
		List<String[]> data = readDataFromCSV(CSVFileName);
		System.out.println("FINISH: reading CSV");
		System.out.println("-----------------------------");
		
		data.remove(0);
		

		Individual district, neighbourhood, chargeStation, equipment, service, connector, operator, geographicalPoint;
		Individual publicCarPark = model.getIndividual(NS+"PublicCarPark");
		Individual publicRoad = model.getIndividual(NS+"PublicRoad");
		Individual excempt = model.getIndividual(NS+"Excempt");
		Individual vECard = model.getIndividual(NS+"VECard");
		Individual eDum = model.getIndividual(NS+"EDum");
		Individual eTaxi = model.getIndividual(NS+"ETaxi");
		Individual motorbike = model.getIndividual(NS+"Motorbike");
		
		// per reutilitzar els individus i no crear-ne de nous cada vegada que apareixin en una fila
		// s'emmagatzemaran en una hastable i abans de crear-ne un de nou, es consultar si ja existeix
		HashMap<String, Individual> districts = new HashMap<String, Individual>();
		HashMap<String, Individual> neighbourhoods = new HashMap<String, Individual>();
		HashMap<String, Individual> chargeStations = new HashMap<String, Individual>();
		HashMap<String, Individual> equipments = new HashMap<String, Individual>();
		HashMap<String, Individual> operators = new HashMap<String, Individual>();
		HashMap<String, Individual> services = new HashMap<String, Individual>();
		HashMap<String, Individual> geographicalPoints = new HashMap<String, Individual>();
		
		// es recuperen les Data Properties definides en l'ontologia
		Property hasCodeId = model.getDatatypeProperty(NS+"hasCodeId");
		Property hasName = model.getDatatypeProperty(NS+"hasName");
		Property hasPostalCode = model.getDatatypeProperty(NS+"hasPostalCode");
		Property hasAddress = model.getDatatypeProperty(NS+"hasAddress");
		Property latitude = model.getDatatypeProperty(geoNS+"lat");
		Property longitude = model.getDatatypeProperty(geoNS+"long");
		Property hasMaxChargePower = model.getDatatypeProperty(NS+"hasMaxChargePower");
		Property hasNumConnectors = model.getDatatypeProperty(NS+"hasNumConnectors");
		Property hasNumSimultaneousServices = model.getDatatypeProperty(NS+"hasNumSimultaneousServices");
		Property hasPrice = model.getDatatypeProperty(NS+"hasPrice");
		Property hasTelephone = model.getDatatypeProperty(NS+"hasTelephone");
		Property asWKT = model.getDatatypeProperty(geosparqlNS+"asWKT");
		
		// es recuperen les Object Properties definides en l'ontologia
		Property isNeighbourhoodOf = model.getProperty(NS+"isNeighbourhoodOf");
		Property isChargeStationOf = model.getProperty(NS+"isChargeStationOf");
		Property hasEquipment = model.getProperty(NS+"hasEquipment");
		Property hasAccessType = model.getProperty(NS+"hasAccessType");
		Property hasAccessControl = model.getProperty(NS+"hasAccessControl");
		Property hasOperator = model.getProperty(NS+"hasOperator");
		Property canOnlyBeUsedBy = model.getProperty(NS+"canOnlyBeUsedBy");
		Property hasService = model.getProperty(NS+"hasService");
		Property hasConnector = model.getProperty(NS+"hasConnector");
		Property geometry = model.getProperty(geoNS+"geometry");

		// es defineix el tipus de dades wktLiteral (non-XSD data type)
		// la classe TypeMapper permet definir un nou tipus de dades a partir d'una URI
		RDFDatatype wktLiteral = TypeMapper.getInstance().getSafeTypeByName("http://www.opengis.net/ont/geosparql#wktLiteral");
		String name;
		
		String lng = "";
		String lat = "";
		
		System.out.println("START: processing CSV...");
		
		// es processa el CSV fila a fila (cada fila cont els diversos camps del CSV)
		for(String[] row: data) {
			
			// es tracta la informaci referent al DISTRICTE
			if(!districts.containsKey(row[13])) {
				district = createIndividual(model, "District", "District_"+row[13]);
			
				model.addLiteral(district, hasCodeId, Integer.parseInt(row[13]));
				district.addProperty(hasName, row[14]);
			
				districts.put(row[13], district);
			}else {
				district = districts.get(row[13]);
			}
			
			
			// es tracta la informaci referent al BARRI
			if(!neighbourhoods.containsKey(row[15])) {
				neighbourhood = createIndividual(model, "Neighbourhood", "Neighbourhood_"+row[15]);
				
				model.addLiteral(neighbourhood, hasCodeId, Integer.parseInt(row[15]));
				neighbourhood.addProperty(hasName, row[16]);
				if(!row[17].equals("")) {
					neighbourhood.addProperty(hasPostalCode, row[17]);
				}
				neighbourhood.addProperty(isNeighbourhoodOf, district);
				
				neighbourhoods.put(row[15], neighbourhood);
			}else {
				neighbourhood = neighbourhoods.get(row[15]);
			}
			
			
			// es tracta la informaci referent a l'OPERADOR
			name = row[20];
			row[20] = row[20].replace(" ","_");
			if(!operators.containsKey(row[20])) {
				if(row[20].contains("Ajuntament")) {
					operator = createIndividual(model, "Council", "Operator_"+row[20]);	
				}else {
					operator = createIndividual(model, "Public", "Operator_"+row[20]);
				}
				
				operator.addProperty(hasName, name);
				operator.addProperty(hasTelephone, row[21]);
						
				operators.put(row[20], operator);
			}else {
				operator = operators.get(row[20]);
			}
			
			
			// es tracta la informaci referent a l'ESTACIO
			if(!chargeStations.containsKey(row[1])) {
				chargeStation = createIndividual(model, "ChargeStation", "ChargeStation_"+row[1]);
				
				model.addLiteral(chargeStation, hasCodeId, Integer.parseInt(row[1]));
				chargeStation.addProperty(isChargeStationOf, neighbourhood);
				
				chargeStations.put(row[1], chargeStation);
			}else {
				chargeStation = chargeStations.get(row[1]);
			}
			
			
			// es tracta la informaci referent al PUNT GEOGRFIC
			if(row[11].contains(",")) {
				// row[11] cont la longitud
				lng = row[11].replace(",", ".");
			}else {
				lng = row[11];
			}
			if(row[10].contains(",")) {
				// row[10] cont la latitud
				lat = row[10].replace(",", ".");
			}else {
				lat = row[10];
			}
			
			if(!geographicalPoints.containsKey(lng+"_"+lat)){
				geographicalPoint = createIndividual(model, geoNS, "Point", "Point_"+lng+"_"+lat);
				
				geographicalPoint.addProperty(asWKT, "POINT("+lng+" "+lat+")", wktLiteral);

				
				geographicalPoints.put(lng+"_"+lat, geographicalPoint);
			}else {
				geographicalPoint = geographicalPoints.get(lng+"_"+lat);
			}
			
			// es tracta la informaci referent a l'EQUIP
			if(!equipments.containsKey(row[1]+row[2])) {
				equipment = createIndividual(model, "Equipment", "Equipment_"+row[1]+row[2]);
				
				model.addLiteral(equipment, hasCodeId, Integer.parseInt(row[2]));
				model.addLiteral(equipment, latitude, Double.parseDouble(lat));
				model.addLiteral(equipment, longitude, Double.parseDouble(lng));
				equipment.addProperty(hasAddress, row[12]);
				
				switch(row[19]) {
					case "VIA PBLICA":
						equipment.addProperty(hasAccessType, publicRoad);
						break;
					case "APARCAMENT PUBLIC":
						equipment.addProperty(hasAccessType, publicCarPark);
				}
				
				switch(row[23]) {
					case "LLIURE":
						equipment.addProperty(hasAccessControl, excempt);
						break;
					case "TARGETA VE":
						equipment.addProperty(hasAccessControl, vECard);
				}
				
				equipment.addProperty(hasOperator, operator);
				equipment.addProperty(geometry, geographicalPoint);
				
				chargeStation.addProperty(hasEquipment, equipment);

				equipments.put(row[1]+row[2], equipment);
			}else {
				equipment = equipments.get(row[1]+row[2]);
			}
			

			// es tracta la informaci referent al SERVEI
			if(!services.containsKey(row[1]+row[2]+row[3])) {
				service = createIndividual(model, "Service", "Service_"+row[1]+row[2]+row[3]);
				
				model.addLiteral(service, hasCodeId, Integer.parseInt(row[3]));
				
				switch(row[22]) {
					case "GRATUT":
						model.addLiteral(service, hasPrice, Float.parseFloat("0"));
				}
				
				switch(row[24]) {
					case "EXCLUSIU eDUM":
						service.addProperty(canOnlyBeUsedBy, eDum);
						break;
					case "EXCLUSIU eTAXI":
						service.addProperty(canOnlyBeUsedBy, eTaxi);
						break;
					case "EXCLUSIU MOTO":
						service.addProperty(canOnlyBeUsedBy, motorbike);
				}
				
				equipment.addProperty(hasService, service);

				services.put(row[1]+row[2]+row[3], service);
			}else {
				service = services.get(row[1]+row[2]+row[3]);
			}
			

			// es tracta la informaci referent al CONNECTOR
			switch(row[5]) {
				case "CHADEMO":
					connector = createIndividual(model, "Chademo", "Chademo_"+row[1]+row[2]+row[3]+row[4]);
					break;
				case "SCHUKO":
					connector = createIndividual(model, "Schuko", "Schuko_"+row[1]+row[2]+row[3]+row[4]);
					break;
				case "MENNEKES":
					connector = createIndividual(model, "Mennekes", "Mennekes_"+row[1]+row[2]+row[3]+row[4]);
					break;
				case "COMBO 2":
					connector = createIndividual(model, "Combo2", "Combo2_"+row[1]+row[2]+row[3]+row[4]);
					break;
				default:
					connector = createIndividual(model, "Connector", "Connector_"+row[1]+row[2]+row[3]+row[4]);
			}
			
			model.addLiteral(connector, hasCodeId, Integer.parseInt(row[4]));
			model.addLiteral(connector, hasMaxChargePower, Integer.parseInt(row[6]));	
			
			model.addLiteral(connector, latitude, Double.parseDouble(lat));
			model.addLiteral(connector, longitude, Double.parseDouble(lng));
			connector.addProperty(geometry, geographicalPoint);
			
			
			service.addProperty(hasConnector, connector);
					
		}
		System.out.println("FINISH: processing CSV...");
		System.out.println("-----------------------------");
		
		System.out.println("START: recalculating 'NUM_SERVEIS_DEL_EQUIP' and 'NUM_DE_CONNECTORS_AL_EQUIP'...");
		
		// es recalcula el nombre de connectors i el nombre de serveis simultanis que t un equip
		int numSimServices;
		int numConnectors;
		StmtIterator it, it2;
		Individual serv;
		
		for(Individual equ: equipments.values()) {
			numSimServices = 0;
			numConnectors = 0;
			it = equ.listProperties(hasService);
			while(it.hasNext()) {
				Statement st = it.next();
				numSimServices++;
				
				String object = st.asTriple().getObject().getLocalName();
				serv = model.getIndividual(NS+object);
				it2 = serv.listProperties(hasConnector);
				while(it2.hasNext()) {
					it2.next();
					numConnectors++;
				}
			}
			model.addLiteral(equ, hasNumConnectors, numConnectors);
			model.addLiteral(equ, hasNumSimultaneousServices, numSimServices);
		}
		System.out.println("FINISH: recalculating 'NUM_SERVEIS_DEL_EQUIP' and 'NUM_DE_CONNECTORS_AL_EQUIP'...");
		System.out.println("-----------------------------");
		
		System.out.println("START: saving the model...");
		save(model);
		System.out.println("FINISH: saving the model...");
		System.out.println("-----------------------------");

	}
	
	
	/**
	 * Mtode que emmagatzema l'ontologia poblada en format RDF, TTL i OWL 
	 * 
	 * @param model ontolgic
	 * 
	 */
	private void save(OntModel m) {
        // Finalment guardem l'ontologia en format RDF   
        File rdf_file = new File("EVCOntology_with_instances.rdf");
		try{
			if (!rdf_file.exists()){
				rdf_file.createNewFile();
			}
			m.writeAll( new OutputStreamWriter( new FileOutputStream(rdf_file), "utf-8"),"RDF/XML");		
		} catch(IOException ioe){
	            ioe.printStackTrace();
		}
		
		// I en format Turtle   
		File ttl_file = new File("EVCOntology_with_instances.ttl");
		try{
			if (!ttl_file.exists()){
				ttl_file.createNewFile();
			}
			m.writeAll( new OutputStreamWriter( new FileOutputStream(ttl_file), "utf-8"),"TTL");		
		} catch(IOException ioe){
	            ioe.printStackTrace();
		}
		
		// I en format OWL   
		File owl_file = new File("EVCOntology_with_instances.owl");
		try{
			if (!owl_file.exists()){
				owl_file.createNewFile();
			}
			m.writeAll( new OutputStreamWriter( new FileOutputStream(owl_file), "utf-8"),"RDF/XML");		
		} catch(IOException ioe){
			    ioe.printStackTrace();
		}		
		
	}
	
	
	/**
	 * Mtode que llegeix el fitxer CSV i emmagatzema el seu contingut en una
	 * llista on cada element es tracta d'un vector de strings
	 * 
	 * @param nom del fitxer CSV que s'ha de llegir
	 * 
	 * @return retorna una llista de vectors de strings on cada vector es correspon a
	 * una fila del CSV i cada posici de cada vector cont els valors de cada camp d'aquella fila
	 * 
	 * @throws IOException 
	 */
	private static List<String[]> readDataFromCSV(String CSVFileName){
		CSVReader csvReader = null;
		List<String[]> data = null;
		
		try {
			csvReader = new CSVReader(new FileReader(CSVFileName));  
			data = csvReader.readAll();
			csvReader.close();
		}catch(IOException e) {
			System.out.println("Error reading CSV file");
			System.exit(0);
		}
		
		return data;
	}
	
	
	/**
	 * Mtode que crea un individu d'una classe
	 * 
	 * @param model ontolgic
	 * @param nom de la classe
	 * @param nom de l'individu a crear
	 * 
	 * @return l'individu que s'ha creat
	 */
	private Individual createIndividual(OntModel model, String className, String instanceName) {
		// es recupera la classe a la qual pertanyer l'individu a crear
		OntClass ontClass = model.getOntClass(NS+className);
		// es crea l'individu
		return ontClass.createIndividual(NS+instanceName);
	}
	
	
	/**
	 * Mtode que crea un individu d'una classe que t un namespace diferent del genric
	 * (el de l'ontologia)
	 * 
	 * @param model ontolgic
	 * @param espai de nom (namespace) 
	 * @param nom de la classe
	 * @param nom de l'individu a crear
	 * 
	 * @return l'individu que s'ha creat
	 */
	private Individual createIndividual(OntModel model, String NS, String className, String instanceName) {
		// es recupera la classe a la qual pertanyer l'individu a crear
		OntClass ontClass = model.getOntClass(NS+className);
		// es crea l'individu
		return ontClass.createIndividual(NS+instanceName);
	}
	
	
}
